/*
  * mesopotamia @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia;


import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.mesopotamia.sql.LoadLevel;
import org.mesopotamia.sql.LoadLevelImpl;

import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.util.ExceptionSink;
import biz.hammurapi.util.StreamPumper;

/**
 * This loader stores zipped file content to the repository without any processing. 
 * 
 * Definition of the loader in database 
<PRE> INSERT INTO LOADER (
     LANGUAGE,
     LANGUAGE_VERSION,
     LEVEL,
     LOADER_CLASS,
     DESCRIPTION,
     REQUIRES_SOURCE,
     REQUIRES_ENVIRONMENT
 ) VALUES (
	'<I>language name</I>',
	'<I>language version</I>',
	'content',
	'org.mesopotamia.ContentLoader',
	'Loads source content into database',
	1,
	0
 );</PRE>
 * @author Pavel
 */
public class ContentLoader extends LoaderBase implements SourceLoader {
	private static final Logger logger = Logger.getLogger(ContentLoader.class.getName());

	public ContentLoader(RepositoryLanguage repoLanguage, LoaderEntry xData) {
		super(repoLanguage, xData);
	}

	public boolean load(final int scanId, final int sourceUnitId, Source source, Object environment) {
		final RepositoryFactory factory=repoLanguage.getFactory();
		try {
			final LoadLevelImpl lli = new LoadLevelImpl(true);
			lli.setLevelId(data.getId());
			lli.setSourceUnitId(sourceUnitId);
			try {
				final boolean[] pumpResult = {true};
				InputStream in = (InputStream) ConvertingService.convert(source.get(), InputStream.class);
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				GZIPOutputStream gzos = new GZIPOutputStream(baos);
				new StreamPumper(
						in, 
						gzos, 
						new ExceptionSink() {

							public void consume(Object source, Exception e) {
								lli.setLoadFailed(true);
								try {
									lli.setMessageId(storeErrorMessage(scanId, sourceUnitId, e));
								} catch (SQLException e1) {
									factory.consume(this, e);
									pumpResult[0] = false;
								}
								logger.warning("Cannot load source '"+source+"' to level "+data.getLevel()+": "+e);
								pumpResult[0] = false;
							}
							
						},
						true).run();
				gzos.close();
				lli.setLoaderData(baos.toByteArray());
				
				return pumpResult[0];
			} catch (Exception e) {
				lli.setLoadFailed(true);
				lli.setMessageId(storeErrorMessage(scanId, sourceUnitId, e));
				logger.warning("Cannot load source '"+source+"' to level "+data.getLevel()+": "+e);
				return false;
			} finally {
				factory.getEngine().insertLoadLevel(lli);
			}
		} catch (SQLException e) {
			factory.consume(this, e);
			return false;
		}
	}
	
	/**
	 * Returns source unit data as input stream.
	 */
	public Object getData(int sourceUnitId, Integer scanId) throws MesopotamiaException {
		RepositoryFactory factory=repoLanguage.getFactory();
		try {
			LoadLevel loadLevel = factory.getEngine().getSourceUnitLoadLevel(data.getId(), sourceUnitId);
			if (loadLevel==null) {
				return null;
			}
			return new GZIPInputStream(new ByteArrayInputStream(loadLevel.getLoaderData()));
		} catch (Exception e) {
			throw new MesopotamiaException("Cannot read tokens: "+e, e);
		}
	}
}
