/*
 * mesopotamia @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia;

import java.awt.event.WindowEvent;
import java.io.StringReader;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.swing.WindowConstants;
import javax.swing.tree.TreeNode;

import org.mesopotamia.util.BnfLexer;
import org.mesopotamia.util.BnfRecognizer;
import org.mesopotamia.util.BnfTokenTypes;
import org.mesopotamia.util.UnmodifiableConvertingList;
import org.w3c.dom.Element;

import antlr.RecognitionException;
import antlr.TokenStreamException;
import antlr.collections.AST;
import biz.hammurapi.convert.ConverterClosure;
import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.legacy.review.Signed;
import biz.hammurapi.swing.Browser;
import biz.hammurapi.swing.Visualizable;
import biz.hammurapi.util.VisitableBase;
import biz.hammurapi.xml.dom.DomSerializable;

/**
 * Generic wrapper for AST. 
 * More sophisticated classes (e.g. try block representation)
 * shall extend this class.
 * @author Pavel Vlasov
 * @revision $Revision: 1.2 $
 */
public class LanguageElement extends VisitableBase implements DomSerializable, Signed, Comparable<LanguageElement> {
	
	private Scan scan;
	private NodeData xData;
	private RepositoryLanguage language;
	protected Object environment;
	private LanguageElementHandle handle;

	/**
	 * Instances of LanguageElement shall be created only by
	 * repository
	 */
	protected LanguageElement(NodeData xData, Class<?> context, Scan scan, RepositoryLanguage language, Object environment) {
		this.scan=scan;
		try {
			this.xData=(NodeData) xData.clone();
		} catch (CloneNotSupportedException e) {
			throw new MesopotamiaRuntimeException(e);
		}
		if (xData instanceof MesopotamiaNode && ((MesopotamiaNode) xData).getParent()!=null) {
			this.setParentHandle(new LanguageElementHandle(xData.getSourceUnitId(), ((MesopotamiaNode) xData).getParent().getId(), null, null)); 
		}
		this.language=language;
		this.environment=environment;
		this.handle=new LanguageElementHandle(xData.getSourceUnitId(), xData.getId(), context, getClass());		
	}
	
	private LanguageElementHandle parentHandle;
	
	public void setParentHandle(LanguageElementHandle parentHandle) {
		if (parentHandle!=null) {
			if (parentHandle.getId()==xData.getId()) {
				throw new IllegalArgumentException("Self parent");
			}
			if (parentHandle.getSourceUnitId()!=xData.getSourceUnitId()) {				
				throw new IllegalArgumentException("Parent from different source unit, this "+xData+", parent "+parentHandle);
			}
		}

		this.parentHandle = parentHandle;
	}
	
	public LanguageElement getParent() {
		return parentHandle==null ?  null : scan.getLanguageElement(parentHandle);
	}
	
	public RepositoryLanguage getLanguage() {
		return language;
	}
	
	public int getType() {
		return language.tokenTypeId2type(xData.getType());
	}
	
	private List<Object> childrenData; // Collection of NodeData
	private List<Object> children;
	
	/**
	 * Language element should not keep references to children, but a 
	 * collection of children's primary keys (java.lang.Integer) and 
	 * use converting collection (biz.hammurapi.util.ConvertingCollection),
	 * which would use Repository's languageElementCache to look up child instance.
	 * @return
	 * @throws MesopotamiaException 
	 */
	public synchronized List<Object> getChildren() throws MesopotamiaException {
		if (childrenData==null) {
			childrenData = new ArrayList<Object>();
			Iterator<MesopotamiaNode> it = getNode().getChildren().iterator();				
			while (it.hasNext()) {
				try {
					childrenData.add(it.next().clone());
				} catch (CloneNotSupportedException e) {
					throw new MesopotamiaRuntimeException(e);
				} 
			}
			
			children=new UnmodifiableConvertingList<Object>(childrenData,data2leConverter); 
		}
		return children;
	}

	public MesopotamiaNode getNode() {
		return getSourceUnit().getSyntaxTree().findNode(xData.getId());
	}
	
	/**
	 * Convenience method
	 * @param idx
	 * @return
	 * @throws MesopotamiaException 
	 */
	public LanguageElement getChild(int idx) throws MesopotamiaException {
		List<Object> ch = getChildren();
		return ch.size()>idx ? (LanguageElement) ch.get(idx) : null;
	}
	
	/**
	 * Convenience method
	 * @param idx
	 * @return
	 * @throws MesopotamiaException 
	 */
	public int getChildId(int idx) throws MesopotamiaException {
		return getChildren().size()>idx ? ((NodeData) childrenData.get(idx)).getId() : 0;
	}
	
	protected ConverterClosure data2leConverter=new ConverterClosure() {

		public Object convert(Object source) {
			NodeData nd = (NodeData) source;
			LanguageElementHandle childHandle = new LanguageElementHandle(nd.getSourceUnitId(), nd.getId(), LanguageElement.this.getClass(), null);
			LanguageElement child = scan.getLanguageElement(childHandle);
			child.setParentHandle(LanguageElement.this.handle);
			return child;
		}
		
	};
	
	protected ConverterClosure le2idConverter=new ConverterClosure() {

		public Object convert(Object source) {
			return new Integer(((LanguageElement) source).getId());
		}
		
	};
	
	public LanguageElement getPrevSibling() {
		return getSibling(-1);
	}

	public LanguageElement getNextSibling() {
		return getSibling(+1);
	}
	
	public LanguageElement getSibling(int offset) {
		SyntaxTree st = getSourceUnit().getSyntaxTree();
		List<MesopotamiaNode> siblings = parentHandle==null ? st.getRoots() : st.findNode(parentHandle.getId()).getChildren();
		int idx = xData.getPosition()+offset;
		if (idx<0 || idx>=siblings.size()) {
			return null;
		}
		NodeData sibling = siblings.get(idx);		
		if (sibling==null) {
			return null;
		}
		LanguageElement parent=getParent();
		LanguageElementHandle siblingHandle=new LanguageElementHandle(
				xData.getSourceUnitId(),
				sibling.getId(),
				parent==null ? null : getParent().getClass(),
				null);
		return scan.getLanguageElement(siblingHandle);
	}

//	public Long getChecksum() {
//		return dbData.getChecksum();
//	}

	public int getCol() {
		return xData.getColumn();
	}

//	public Token getFirstToken() {
//		return xData.getFirstTokenId()==null ? null : getSourceUnit().getToken(dbData.getFirstTokenId().intValue());
//	}

	public int getId() {
		return xData.getId();
	}

//	public Token getLastToken() {
//		return dbData.getLastTokenId()==null ? null : getSourceUnit().getToken(dbData.getLastTokenId().intValue());
//	}
//	
//	public Token getToken() {
//		return dbData.getTokenId()==null ? null : getSourceUnit().getToken(dbData.getTokenId().intValue());
//	}

	public int getLine() {
		return xData.getLine();
	}

	/**
	 * Change to lazy calculation once in the future.
	 * @return
	 */
	public String getSignature() {
		StringBuffer ret=new StringBuffer();
		LanguageElement parent = getParent();
		if (parent==this) {
			/*
			 * There might be cases where signature is not applicable. 
			 * We use element ID and location in this case. This signature is not as good as
			 * regular signature, but there is no better way.
			 */ 			
			return getSourceUnit().getPath()+":ID:"+getId(); 
		}
		if (parent==null) {
			ret.append(getSourceUnit().getPath());
		} else {
			ret.append(parent.getSignature());
		}
		ret.append(":");
		ret.append(language.tokenTypeId2name(xData.getType()));
//		if (xData.getSignatureId()!=null) {
//			ret.append("[");
//			ret.append(getText(dbData.getSignatureId()));
//			ret.append("]");
//		} else {
			if (xData.getSameTypeIndex()!=0) {
				ret.append("[");
				ret.append(xData.getSameTypeIndex());
				ret.append("]");
			}
//		}
		return ret.toString();
	}

//	/**
//	 * Retrieves text from string storage
//	 * @return
//	 */
//	protected String getText(Integer id) {		
//		return scan.getRepository().getFactory().getStringStorage().getText(id);
//	}

	public AstSourceUnit getSourceUnit() {
		return (AstSourceUnit) scan.getSourceUnit(xData.getSourceUnitId());
	}

	/**
	 * @return Language element text
	 */
	public String getText() {
		return xData.getText(); //getText(dbData.getTextId());
	}

//	public Token getToken() throws MesopotamiaException {
//		// It would be too cumbersome to keep tokens in memory cache - instantiate them
//		// directly. It can be changed in the future anyway.
//		if (dbData.getTokenId()==null) {
//			return null;
//		}
//		
//		try {
//			return new Token(getSourceUnit(), repository.getFactory().getEngine().getToken(dbData.getSourceUnitId(), dbData.getTokenId().intValue()));
//		} catch (SQLException e) {
//			throw new MesopotamiaException(e);
//		} 
//	}

//	public int getSize() {
//		return dbData.getTotalSize();
//	}

	public void toDom(Element holder) {
		holder.setAttribute("id",String.valueOf(xData.getId()));
		holder.setAttribute("column",String.valueOf(xData.getColumn()));
		holder.setAttribute("line",String.valueOf(xData.getLine()));
		holder.setAttribute("token-type",language.tokenTypeId2name(xData.getType()));
		holder.setAttribute("class",getClass().getName());
		setAttribute(holder, "text",getText());
//		Token token = getToken();
//		if (token!=null) {
//			holder.setAttribute("token-position", String.valueOf(token.getPosition()));
//		}
	}

	/**
	 * Token name, e.g. METHOD_CALL
	 * @return
	 */
	public String getTokenName() {
		return language.tokenTypeId2name(xData.getType());
	}	
	
//	protected org.mesopotamia.sql.LanguageElement[] getDbChildrenByType(int type) throws MesopotamiaException {		
//		getChildren();
//		Collection ret=new ArrayList();
//		Iterator it=childrenIds.iterator();
//		int typeId=language.tokenType2id(type);
//		while (it.hasNext()) {
//			org.mesopotamia.MesopotamiaNodedbLe=(org.mesopotamia.sql.LanguageElement) it.next();
//			if (dbLe.getTokenTypeId()==typeId) {
//				ret.add(dbLe);
//			}
//		}
//		return (org.mesopotamia.sql.LanguageElement[]) ret.toArray(new org.mesopotamia.sql.LanguageElement[ret.size()]);
//	}
	
	protected Scan getScan() {
		return scan;
	}
	
	/**
	 * Selects collection of language elements based on path expression.
	 * @param context
	 * @param path
	 * @return List of language elements
	 */
	public <T> UnmodifiableConvertingList<T> select(final Class<T> targetClass, String path) throws MesopotamiaException {
		try {
			StringReader sr=new StringReader(path);
			BnfLexer lexer=new BnfLexer(sr);
			BnfRecognizer parser=new BnfRecognizer(lexer);
			parser.standaloneIndexedPath();
			List<NodeData> results=new ArrayList<NodeData>();
			//System.out.println("=== Selecting: "+path);
			//AstUtil.dump(parser.getAST(), parser.getTokenNames());
			for (AST pe=parser.getAST(); pe!=null; pe=pe.getNextSibling()) {
				List<NodeData> peResults=new ArrayList<NodeData>();
				peResults.add(xData);
				for (AST node=pe.getFirstChild(); node!=null; node=node.getNextSibling()) {
					peResults=processNode(peResults, node.getFirstChild());
				}
				results.addAll(peResults);
			}
			
			return new UnmodifiableConvertingList<T>(
					results,
					new ConverterClosure() {

						public Object convert(Object source) {
							NodeData nd = (NodeData) source;
							LanguageElementHandle handle = new LanguageElementHandle(nd.getSourceUnitId(), nd.getId(), LanguageElement.this.getClass(), targetClass);
							LanguageElement selectedChild = scan.getLanguageElement(handle);
							selectedChild.setParentHandle(LanguageElement.this.handle);
							return selectedChild;
						}
						
					});

		} catch (RecognitionException e) {
			throw new MesopotamiaException("Cannot parse path: "+path, e);
		} catch (TokenStreamException e) {
			throw new MesopotamiaException("Cannot parse path: "+path, e);
		} catch (ClassNotFoundException e) {
			throw new MesopotamiaException("Cannot process selection from path: "+path, e);
		}
	}
	
	private List<NodeData> processNode(List<NodeData> input, AST node) throws MesopotamiaException, ClassNotFoundException {
		//System.out.println("--- Processing: "+BnfRecognizer._tokenNames[node.getType()]);
		SyntaxTree st = getSourceUnit().getSyntaxTree();
		List<NodeData> ret=new ArrayList<NodeData>();
		Iterator<NodeData> it=input.iterator();
		while (it.hasNext()) {
			List<NodeData> receiver=new ArrayList<NodeData>();
			MesopotamiaNode contextNode = st.findNode(it.next().getId());
			switch (node.getType()) {
			case BnfTokenTypes.IDENT: {
				Iterator<MesopotamiaNode> cit = contextNode.getChildren().iterator();
				while (cit.hasNext()) {
					NodeData cnd = cit.next();
					checkSameSourceUnit(cnd);
					if (cnd.getType()==language.tokenName2id(node.getText())) {
						receiver.add(cnd);
					}
				}
				break;
			}
			case BnfTokenTypes.EXCL: {
				Iterator<MesopotamiaNode> cit = contextNode.getChildren().iterator();
				while (cit.hasNext()) {
					NodeData cnd = cit.next();
					checkSameSourceUnit(cnd);
					if (cnd.getType()!=language.tokenName2id(node.getFirstChild().getText())) {
						receiver.add(cnd);
					}
				}
				break;
			}
			case BnfTokenTypes.STAR:
				receiver.addAll(contextNode.getChildren());
				break;
			case BnfTokenTypes.DOUBLEDOT:
				throw new UnsupportedOperationException("Not yet implemented");
			case BnfTokenTypes.TYPE_REF:
				StringBuffer typeName=new StringBuffer();
				typeName(node.getFirstChild(), typeName);
				Class<?> type=language.getClassLoader().loadClass(typeName.toString());
				Iterator<MesopotamiaNode> lit=contextNode.getChildren().iterator();
				while (lit.hasNext()) {
					NodeData cnd = lit.next();
					checkSameSourceUnit(cnd);
					LanguageElement cle=scan.getLanguageElement(new LanguageElementHandle(cnd.getSourceUnitId(), cnd.getId(), this.getClass(), null));
					if (type.isInstance(cle)) {
						receiver.add(cnd);
					}
				}
				
				break;
			default:
				throw new MesopotamiaException("Unexpected node type: "+BnfRecognizer._tokenNames[node.getType()]);					
			}
			
			if (node.getNextSibling()!=null) {
				int idx=Integer.parseInt(node.getNextSibling().getFirstChild().getText());
				if (idx>=0 && idx<receiver.size()) {
					ret.add(receiver.get(idx));
				}
			} else {
				ret.addAll(receiver);
			}
		}
		return ret;
	}
	
	private void checkSameSourceUnit(NodeData cnd) {
		if (cnd.getSourceUnitId()!=xData.getSourceUnitId()) {
			throw new IllegalStateException("Nodes from different source units: "+xData+", "+cnd);
		}
		
	}

	static void typeName(AST node, StringBuffer sb) {
		if (node.getType()==BnfTokenTypes.DOT) {
			typeName(node.getFirstChild(), sb);
			sb.append(".");
			typeName(node.getFirstChild().getNextSibling(), sb);
		} else {
			sb.append(node.getText());
		}
	}
	
	/**
	 * Selects single language element handle based on path expression
	 * @param context
	 * @param path
	 * @return
	 * @throws MesopotamiaException 
	 */
	public LanguageElementHandle selectSingleElementHandle(Class targetClass, String path) throws MesopotamiaException {
		UnmodifiableConvertingList<LanguageElement> c=select(targetClass, path);
		if (c.isEmpty()) {
			return null;
		}
		NodeData nd = (NodeData) c.getMaster().get(0);
		return new LanguageElementHandle(nd.getSourceUnitId(), nd.getId(), getClass(), targetClass);		
	}
	
	/**
	 * Selects single language element based on path expression
	 * @param context
	 * @param path
	 * @return
	 * @throws MesopotamiaException 
	 */
	public <T> T selectSingleElement(Class<T> targetClass, String path) throws MesopotamiaException {
		List<T> c=select(targetClass, path);
		return (T) (c.isEmpty() ? null : c.get(0));		
	}
	
	public <T> String selectSingleElementText(Class<T> targetClass, String path) throws MesopotamiaException {
		T le=selectSingleElement(targetClass, path);
		return le instanceof LanguageElement ? ((LanguageElement) le).getText() : null;
	}
	
	public List<String> selectText(Class targetClass, String path) throws MesopotamiaException {
		List<String> ret=new ArrayList<String>();
		Iterator<LanguageElement> it=select(targetClass, path).iterator();
		while (it.hasNext()) {
			ret.add(it.next().getText());
		}
		return ret;
	}	
	
	/**
	 * Sets attribute if value is not null.
	 * @param holder
	 * @param name
	 * @param value
	 */
	protected void setAttribute(Element holder, String name, String value) {
		if (value!=null) {
			holder.setAttribute(name, value);
		}
	}
	
	/**
	 * Adds nested element if value is not null
	 * @param holder
	 * @param name
	 * @param value
	 */
	protected void setElement(Element holder, String name, Object value) {
		if (value!=null) {
			Element el=holder.getOwnerDocument().createElement(name);
			holder.appendChild(el);
			DomSerializable ds = (DomSerializable) ConvertingService.convert(value, DomSerializable.class);
			ds.toDom(el);
		}
	}
	
	/**
	 * Creates a converting list of language elements. Use this method to construct
	 * custom lists of language element children.
	 * @param idList
	 * @param targetType
	 * @return
	 */
	protected List<LanguageElement> createLanguageElementsList(List<Object> nodeDataList, final Class targetType) {
		return new UnmodifiableConvertingList<LanguageElement>(
				nodeDataList,
				new ConverterClosure() {

					public Object convert(Object source) {
						NodeData nd = (NodeData) source;						
						LanguageElementHandle leHandle = new LanguageElementHandle(
								nd.getSourceUnitId(), 
								nd.getId(), 
								LanguageElement.this.getClass(), 
								targetType);
						
						LanguageElement ret = getScan().getLanguageElement(leHandle);
						ret.setParentHandle(handle);
						return ret;
					}
			
		});
	}
	
	/**
	 * @return Language element handle.
	 */
	public LanguageElementHandle getHandle() {
		return handle;
	}

	/**
	 * For sorting.
	 */
	public int compareTo(LanguageElement otherElement) {
		int pathComparison=getSourceUnit().getPath().compareTo(otherElement.getSourceUnit().getPath());
		if (pathComparison==0) {
			int lineComparison=getLine()-otherElement.getLine();
			if (lineComparison==0) {
				return getCol()-otherElement.getCol();
			}
			return lineComparison;
		}
		return pathComparison;
	}
	
	public String getLocation() {
		return getSourceUnit().getPath()+" "+getLine()+":"+getCol();
	}
	
	public String toString() {
		return "["+getClass().getName()+"] "+getTokenName()+" '"+getText()+"' "+getLocation();
	}
	
	// Empty list to initialize list attributes
	@SuppressWarnings("unchecked")
	protected static final UnmodifiableConvertingList emptyList=
		new UnmodifiableConvertingList(Collections.unmodifiableList(new ArrayList()), null);
	
	/**
	 * Shows element in browser. Use it for debugging.
	 */
	public void show() {
		Visualizable vs = (Visualizable) ConvertingService.convert(this, Visualizable.class); 
		String nName = getClass().getName();
		int idx = nName.lastIndexOf(".");
		String cName = idx==-1 ? nName : nName.substring(idx+1);
		TreeNode root = vs.toTreeNode(null, cName);
		
		final Object monitor = new Object();
		
		Browser browser = new Browser(cName+" "+getId()+" at "+getLocation(), root) {
			
			protected void processWindowEvent(WindowEvent e) {
				super.processWindowEvent(e);
				if (e.getID()==WindowEvent.WINDOW_CLOSED) {
					synchronized (monitor) {
						monitor.notifyAll();
					}
				}
			}
		};
		
		browser.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		browser.setVisible(true);
		
		synchronized (monitor) {
			try {
				monitor.wait();
			} catch (InterruptedException ex) {
				// Ignore
			}
		}
	}
	
	/**
	 * Navigates up the tree looking for parent of particular type.
	 * @param clazz Type of parent to be found.
	 * @return Parent of type specified in <code>clazz</code> parameter.
	 */
	@SuppressWarnings("unchecked")
	public <T> T findParent(Class<T> clazz) {
		LanguageElement parent = getParent();

		if (clazz.isInstance(parent) && parent != this && parent.getId()!=this.getId()) {
			return (T) parent;
		}

		if (parent!=null && parent != this && parent.getId()!=this.getId()) {
			LanguageElement ret = (LanguageElement) parent.findParent(clazz);
			if (ret!=null) {
				return (T) ret;
			}
		}

		// Finding using syntax tree
		for (MesopotamiaNode parentNode = getNode().getParent(); parentNode!=null && parentNode.getParent()!=parentNode; parentNode = parentNode.getParent()) {
			LanguageElementHandle lh=new LanguageElementHandle(parentNode.getSourceUnitId(), parentNode.getId(), null, null);
			LanguageElement pe = scan.getLanguageElement(lh);
			if (clazz.isInstance(pe) && pe != this && pe.getId()!=this.getId()) {
				return (T) pe;
			}
		}
		return null;
	}
	
	/**
	 * Override it to return false for methods which shall not be 
	 * mounted to visualizer tree, i.e. methods which return references 
	 * to other language elements. 
	 * @return true if method return value shall be shown in the visualizer tree.
	 */
	public boolean showInTree(Method method) {
		return method!=null && !method.getDeclaringClass().equals(LanguageElement.class) && LanguageElement.class.isAssignableFrom(method.getDeclaringClass());
	}
	
}
