/*
 * mesopotamia @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia;

import gnu.trove.TIntObjectHashMap;

import java.lang.ref.Reference;
import java.lang.ref.SoftReference;
import java.lang.reflect.InvocationTargetException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.mesopotamia.sql.ErrorMessageImpl;
import org.mesopotamia.sql.LanguageImpl;
import org.mesopotamia.sql.MesopotamiaEngine;
import org.mesopotamia.sql.RepositoryImpl;

import biz.hammurapi.config.Context;
import biz.hammurapi.convert.ConverterClosure;
import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.persistence.StringStorage;
import biz.hammurapi.sql.IdentityGenerator;
import biz.hammurapi.sql.IdentityManager;
import biz.hammurapi.sql.IdentityRetriever;
import biz.hammurapi.sql.RowProcessor;
import biz.hammurapi.sql.SQLExceptionEx;
import biz.hammurapi.sql.SQLProcessor;
import biz.hammurapi.sql.Transaction;
import biz.hammurapi.util.ExceptionSink;
import biz.hammurapi.util.Worker;


/**
 * Central class which is responsible for
 * managing languages and repositories.
 * @author Pavel Vlasov
 * @revision $Revision: 1.2 $
 */
public class RepositoryFactory implements ExceptionSink {	
//	private static final MeasurementCategory mc=MeasurementCategoryFactory.getCategory(RepositoryFactory.class);
	private static final Logger logger = Logger.getLogger(RepositoryFactory.class.getName());
	
	final ConverterClosure toIntegerConverter=new ConverterClosure() {

		public Object convert(Object source) {
			return ConvertingService.convert(source, Integer.class);
		}
		
	};
	
	/**
	 * Converts "LEVEL_ID" column obtained from context into integer
	 */
	final ConverterClosure levelIdToIntegerConverter=new ConverterClosure() {

		public Object convert(Object source) {
			if (source instanceof Context) {
				return ConvertingService.convert(((Context) source).get("LEVEL_ID"), Integer.class);
			} 
			return null;
		}
		
	};
	
	int chunkSize=250; // TODO - read from configuration table.
	
	/**
	 * Responsible for retrieving unique keys from the database.
	 */
	private IdentityManager identityManager;
	
	/**
	 * Used in engine constructor and to perform transactional operations.
	 */
	private SQLProcessor processor;
	
	/**
	 * Primary object to access the database.
	 */
	private MesopotamiaEngine engine;
	
	private Map<String, Object> globalParameters=new HashMap<String, Object>();

	private Timer timer;

	private ClassLoader classLoader;

	public Object getGlobalParameter(String name) {
		return globalParameters.get(name);
	}
		
	/**
	 * @param processor
	 * @param identityManager
	 * @throws SQLException
	 * @param timer Timer to schedule cache cleaning tasks. If null then internal timer is created.
	 */
	public RepositoryFactory(
			SQLProcessor processor, 
			Worker worker, 
			Timer timer, 
			StringStorage stringStorage, 
			ClassLoader clsLoader) throws SQLException {
		
		super();		
		this.processor = processor;
		this.engine=new MesopotamiaEngine(processor);
		this.worker=worker;
		this.timer = timer;
		this.classLoader = clsLoader==null ? getClass().getClassLoader() : clsLoader;
		if (timer==null) {
			this.timer = new Timer(true);
		}
		
		this.stringStorage = stringStorage;
		
		engine.processGlobalParameter(
				new RowProcessor() {

					public boolean process(ResultSet rs) throws SQLException {
						String value=rs.getString("PARAMETER_VALUE");
						try {
							if (value==null || value.trim().length()==0) {
									globalParameters.put(
											rs.getString("NAME"), 
											classLoader.loadClass(rs.getString("CLASS_NAME")).newInstance());
							} else {
								globalParameters.put(
										rs.getString("NAME"), 
										classLoader.loadClass(rs.getString("CLASS_NAME"))
											.getConstructor(new Class[] {String.class})
											.newInstance(new Object[] {value}));
								
							}
						} catch (InstantiationException e) {
							throw new SQLExceptionEx("Cannot instantiate parameter '"+rs.getString("NAME")+"'", e);
						} catch (IllegalAccessException e) {
							throw new SQLExceptionEx("Cannot instantiate parameter '"+rs.getString("NAME")+"'", e);
						} catch (ClassNotFoundException e) {
							throw new SQLExceptionEx("Cannot instantiate parameter '"+rs.getString("NAME")+"'", e);
						} catch (SecurityException e) {
							throw new SQLExceptionEx("Cannot instantiate parameter '"+rs.getString("NAME")+"'", e);
						} catch (SQLException e) {
							throw new SQLExceptionEx("Cannot instantiate parameter '"+rs.getString("NAME")+"'", e);
						} catch (InvocationTargetException e) {
							throw new SQLExceptionEx("Cannot instantiate parameter '"+rs.getString("NAME")+"'", e);
						} catch (NoSuchMethodException e) {
							throw new SQLExceptionEx("Cannot instantiate parameter '"+rs.getString("NAME")+"'", e);
						}
						return true;
					}
					
				});
		
		
		this.identityManager = (IdentityManager) getGlobalParameter("IdentityManager");
		this.chunkSize = ((Integer) getGlobalParameter("SharedTextChunkSize")).intValue();
		
		engine.processLanguage(
				new RowProcessor() {

					public boolean process(ResultSet rs) throws SQLException {
						LanguageImpl li=new LanguageImpl(rs);
						languages.put(new Language(li), new RepositoryLanguage(RepositoryFactory.this, li, classLoader));
						return true;
					}
					
				});							
	}
	
	public MesopotamiaEngine getEngine() {
		return engine;
	}

	public IdentityManager getIdentityManager() {
		return identityManager;
	}

	public SQLProcessor getProcessor() {
		return processor;
	}

	public Repository createRepository(String name) throws MesopotamiaException {
		try {
			final RepositoryImpl ri=new RepositoryImpl(true);
			ri.setName(name);
			processor.executeTransaction(new Transaction() {
	
				public boolean execute(SQLProcessor processor) throws SQLException {					
					if (identityManager instanceof IdentityGenerator) {
						ri.setId(((IdentityGenerator) identityManager).generate(processor.getConnection(), "REPOSITORY"));
					}
					new MesopotamiaEngine(processor).insertRepository(ri);
					if (identityManager instanceof IdentityRetriever) {
						ri.setId(((IdentityRetriever) identityManager).retrieve(processor.getConnection()));
					}
					return true;
				}
				
			});
			return new Repository(this, ri);
		} catch (SQLException e) {
			throw new MesopotamiaException("Cannot create repository", e);
		}
	}
	
	public Repository getRepository(int id) throws MesopotamiaException {
		synchronized (repositoryMap) {
			Repository ret=(Repository) repositoryMap.get(id);
		
			if (ret!=null) {
				return ret;
			}
			
			try {			
				org.mesopotamia.sql.Repository ri=engine.getRepository(id);
				if (ri==null) {
					throw new MesopotamiaException("Invalid repository id: " + id);
				}
				
				ret = new Repository(this, ri);
				repositoryMap.put(ret.getId(), ret);
				return ret;
			} catch (SQLException e) {
				throw new MesopotamiaException("Cannot load repository", e);
			}
		}
	}
	
	private Map<Language, RepositoryLanguage> languages=new HashMap<Language, RepositoryLanguage>();
	
	public RepositoryLanguage getRepositoryLanguage(Language language) {
		return languages.get(language);
	}
	
	public Collection<RepositoryLanguage> getRepositoryLanguages() {
		return Collections.unmodifiableCollection(languages.values());
	}

	
	/**
	 * Returns message digest by name.
	 * @param algorithm
	 * @return
	 * @throws MesopotamiaException 
	 */
	public MessageDigest getMessageDigest(String algorithm) throws MesopotamiaException {
		try {
			MessageDigest ret = MessageDigest.getInstance(algorithm);
			return ret;
		} catch (NoSuchAlgorithmException e) {
			throw new MesopotamiaException("Invalid digest algorithm: "+algorithm, e);
		}
	}
	
	RepositoryLanguage getRepositoryLanguageByTokenTypeId(int tokenTypeId) {
		Iterator<RepositoryLanguage> lit=languages.values().iterator();
		while (lit.hasNext()) {
			RepositoryLanguage rl=lit.next();
			if (rl.belongsTo(tokenTypeId)) {
				return rl;
			}
		}
		return null;
	}
	
	private Worker worker;
	
	/**
	 * Processes job
	 * @param job
	 */
	public void process(Runnable job) {
		if (job!=null) {
			if (job instanceof MesopotamiaJob) {
				((MesopotamiaJob) job).setFactory(this);
			}
			
			if (worker==null || !worker.post(job)) {
				job.run();
			}
		}
	}

	/**
	 * Override if needed.
	 */
	public void consume(Object source, Exception e) {
		logger.log(Level.SEVERE, "Exception in "+source+": "+e, e);
	}
	
	private class SoftIntObjectMap extends TIntObjectHashMap {
		/**
		 * 
		 */
		private static final long serialVersionUID = 5449550690446482445L;

		public Object put(int key, Object value) {
			return super.put(key, new SoftReference<Object>(value));
		}
		
		public Object get(int key) {
			Reference<?> ref = (Reference<?>) super.get(key);
			if (ref==null) {
				return null;
			}
			
			Object obj = ref.get();
			if (obj==null) {
				remove(key);
			}
			
			return obj; 
		}
	}
	
	private SoftIntObjectMap scanMap=new SoftIntObjectMap();
	
	void addScan(Scan scan) {
		synchronized (scanMap) {
			scanMap.put(scan.getId(), scan);
		}
	}
	
	public Scan getScan(int id) throws MesopotamiaException {
		synchronized (scanMap) {
			Scan ret = (Scan) scanMap.get(id);
		
			if (ret!=null) {
				return ret;
			}
			
			try {
				org.mesopotamia.sql.Scan si=getEngine().getScan(id);
				if (si==null) {
					return null;
				}
				
				Repository repo=getRepository(si.getRepository());
				ret=new Scan(repo, si);
				addScan(ret);
				return ret;
			} catch (SQLException e) {
				throw new MesopotamiaException("Cannot load scan", e);
			}
		}		
	}
	
	private SoftIntObjectMap repositoryMap=new SoftIntObjectMap();

	private StringStorage stringStorage;	
	
	public StringStorage getStringStorage() {
		return stringStorage;
	}

	Timer getTimer() {
		return timer;
	}
	
	/**
	 * Helper method for loaders.
	 * @param scanId
	 * @param sourceUnitId
	 * @param errorType
	 * @param errorMessage
	 * @return Message record ID.
	 * @throws SQLException
	 */
	public int storeErrorMessage(final int scanId, final Integer sourceUnitId, final String errorType, final String errorMessage) {
		try {
			final ErrorMessageImpl emi=new ErrorMessageImpl(true);
			getProcessor().executeTransaction(
					new Transaction() {
		
						public boolean execute(SQLProcessor processor) throws SQLException {
							Connection con = processor.getConnection();
							IdentityManager identityManager = getIdentityManager();
							if (identityManager instanceof IdentityGenerator) {
								emi.setId(((IdentityGenerator) identityManager).generate(con, "ERROR_MESSAGE"));
							}
							
							emi.setErrorType(errorType);
							emi.setMessageText(errorMessage);
							emi.setSourceUnitId(sourceUnitId);
							emi.setScanId(new Integer(scanId));
							
							new MesopotamiaEngine(processor).insertErrorMessage(emi);
							
							if (identityManager instanceof IdentityRetriever) {
								emi.setId(((IdentityRetriever) identityManager).retrieve(con));
							}
							return true;
						}
						
					});
			return emi.getId();
		} catch (Exception e) {
			consume(this, e);
			return -1;
		}
	}
	
	public int storeErrorMessage(final int scanId, final Integer sourceUnitId, final Throwable e) {
		Throwable rootCause=e;
		while (rootCause.getCause()!=null) {
			rootCause=rootCause.getCause();
		}
		
		return storeErrorMessage(
			scanId, 
			sourceUnitId, 
			rootCause.getClass().getName(), 
			""+rootCause.getMessage());
		
	}

	/**
	 * Deletes source unit from the database.
	 * @param id Source unit ID.
	 * @throws MesopotamiaException
	 */
	public void deleteSourceUnit(int id) throws MesopotamiaException {
		try {
			engine.deleteSourceUnitScanBySourceUnit(id);
			engine.deleteSourceUnit(id);
		} catch (SQLException e) {
			throw new MesopotamiaException(e);
		}
	}
	
}
