/*
 * mesopotamia @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia;

import gnu.trove.TIntHashSet;
import gnu.trove.TIntIntHashMap;
import gnu.trove.TIntObjectHashMap;
import gnu.trove.TObjectIntHashMap;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.mesopotamia.sql.LanguageElementClassImpl;
import org.mesopotamia.sql.SourceUnitClassImpl;
import org.mesopotamia.sql.TokenType;

import biz.hammurapi.sql.RowProcessor;
import biz.hammurapi.sql.SQLExceptionEx;

/**
 * Language defined in repository.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class RepositoryLanguage extends Language {
	private static final Logger logger = Logger.getLogger(RepositoryLanguage.class.getName());

	private RepositoryFactory factory;
	private ClassLoader classLoader;
	
	public ClassLoader getClassLoader() {
		return classLoader;
	}

//	public RepositoryLanguage(String name, String version, String description) {
//		super(name, version, description);
//		// TODO Auto-generated constructor stub
//	}

	public RepositoryLanguage(final RepositoryFactory factory, org.mesopotamia.sql.Language data, ClassLoader classLoader) throws SQLException {
		super(data);
		this.factory=factory;
		this.classLoader=classLoader;
		// TODO Read path elements from database table and construct
		// TODO classloader if there are any. It will allow to keep language module
		// TODO jars centrally.
		
		Iterator it=factory
			.getEngine()
			.getTokenTypeByLanguage(
					dbData.getName(), 
					dbData.getVersion())
				.iterator();
	
		while (it.hasNext()) {
			TokenType tt = (org.mesopotamia.sql.TokenType) it.next();
			type2id.put(tt.getTokenType(), tt.getId());
			id2type.put(tt.getId(), tt.getTokenType());
			name2id.put(tt.getTokenName(), tt.getId());
			id2name.put(tt.getId(), tt.getTokenName());
			if (tt.getIsWhitespace()) {
				whitespaces.add(tt.getId());
			}
			
			final Collection<LanguageElementClassEntry> lecs=new ArrayList<LanguageElementClassEntry>();
			factory.getEngine().processLanguageElementClassByTokenType(
					tt.getId(),
					new RowProcessor() {
						public boolean process(ResultSet rs) throws SQLException {
							lecs.add(new LanguageElementClassEntry(rs));
							return true;
						}
						
					});		
			languageElementClasses.put(tt.getId(), lecs);
		}
		
		loadLoaders();
		
		factory.getEngine().processSourceUnitClassByLanguage(
				getName(),
				getVersion(),
				new RowProcessor() {

					public boolean process(ResultSet rs) throws SQLException {
						sourceUnitClasses.add(new SourceUnitClassEntry(rs));
						return true;
					}
					
				});		
	}
	
	private List<LoaderEntry> loaders;
	
	/**
	 * @return loaders ordered by dependency 
	 */
	List<LoaderEntry> getLoaders() {
		return loaders;
	}
	
	private Map<String, Loader> levelMap = new HashMap<String, Loader>();
	
	public Loader getLoader(String level) {
		return levelMap.get(level);
	}
	
	TIntObjectHashMap loadersMap=new TIntObjectHashMap();
	
	/**
	 * @param language
	 * @return List of loaders ordered by dependency
	 * @throws SQLException 
	 * @throws MesopotamiaException If loaders cannot be retrieved from the database or there is a circular dependency between
	 * loaders
	 */
	private void loadLoaders() throws SQLException {
		final List<LoaderEntry> ldrs=new ArrayList<LoaderEntry>();
		factory.getEngine().processLoaderByLanguage(
				getName(), 
				getVersion(), 
				new RowProcessor() {

					public boolean process(ResultSet rs) throws SQLException {
						LoaderEntry loaderEntry = new LoaderEntry(RepositoryLanguage.this, rs);
						ldrs.add(loaderEntry);
						loadersMap.put(loaderEntry.getId(), loaderEntry);
						levelMap.put(loaderEntry.getLevel(), loaderEntry.getLoader());
						return true;
					}
					
				});
		
		// Sorting loaders by dependency
		Collections.sort(
				ldrs, 
				new Comparator<LoaderEntry>() {

					public int compare(LoaderEntry o1, LoaderEntry o2) {
						if (o1==o2) {
							return 0;
						}
						
						LoaderEntry le1=(LoaderEntry) o1;
						LoaderEntry le2=(LoaderEntry) o2;
						if (le1.dependsOn(le2.getId(), le2.getId())) {
							return 1;
						}
						if (le2.dependsOn(le1.getId(), le1.getId())) {
							return -1;
						}
						return le1.getId()-le2.getId();								
					}
			
		});

		loaders=Collections.unmodifiableList(ldrs);
	}
	
	public RepositoryFactory getFactory() {
		return factory;
	}
	
	private TIntIntHashMap type2id=new TIntIntHashMap();
	private TIntIntHashMap id2type=new TIntIntHashMap();
	private TObjectIntHashMap name2id=new TObjectIntHashMap();
	private TIntObjectHashMap id2name=new TIntObjectHashMap();
	private TIntHashSet whitespaces = new TIntHashSet();
	
	/**
	 * Converts token type as defined in grammar to token type id as stored in repository
	 * @param tokenType
	 * @return
	 */
	public int tokenType2id(int tokenType) {
		return type2id.get(tokenType);
	}
	
	public boolean isWhitespace(int typeId) {
		return whitespaces.contains(typeId);
	}
	
	/**
	 * Converts token type id as defined in repository to token type as defined in grammar.
	 * @param tokenTypeId
	 * @return
	 */
	public int tokenTypeId2type(int tokenTypeId) {
		return id2type.get(tokenTypeId);
	}
	
	public int tokenName2id(String tokenName) throws MesopotamiaException {
		if (supportsTokenName(tokenName)) {
			return name2id.get(tokenName);
		}
		throw new MesopotamiaException("Invalid token name: "+tokenName);
	}

	/**
	 * @param tokenName
	 * @return
	 */
	public boolean supportsTokenName(String tokenName) {
		return name2id.containsKey(tokenName);
	}
	
	/**
	 * Converts token type id as defined in repository to token type as defined in grammar.
	 * @param tokenTypeId
	 * @return
	 */
	public String tokenTypeId2name(int tokenTypeId) {
		return (String) id2name.get(tokenTypeId);
	}
	
	private class SourceUnitClassEntry extends SourceUnitClassImpl implements Comparable {
		private Set requiredLevels=new HashSet();
		private Class sourceUnitClass;
		
		public SourceUnitClassEntry(ResultSet rs) throws SQLException {
			super(rs);
			try {
				sourceUnitClass=classLoader.loadClass(getClassName());
			} catch (ClassNotFoundException e) {
				throw new SQLExceptionEx("Cannot load source unit class "+getClassName(), e);
			}
			if (!SourceUnit.class.isAssignableFrom(sourceUnitClass)) {
				throw new SQLException(sourceUnitClass.getName()+" must be a subclass of "+SourceUnit.class.getName());
			}
			factory.getEngine().getSourceUnitClassRequiredLevels(
					getId(), 
					requiredLevels,
					factory.toIntegerConverter);
		}
		
		boolean isCompatible(Collection loadedLevels) {
			return loadedLevels.containsAll(requiredLevels);
		}

		public int compareTo(Object o) {
			if (o==this) {
				return 0;
			}
			
			if (o instanceof SourceUnitClassEntry) {
				SourceUnitClassEntry s = (SourceUnitClassEntry) o;
				if (sourceUnitClass.equals(s.sourceUnitClass)) {
					throw new IllegalStateException("Two entries with the same class "+sourceUnitClass.getName());
				}
				
				if (sourceUnitClass.isAssignableFrom(s.sourceUnitClass)) {
					return 1;
				}
				
				if (s.sourceUnitClass.isAssignableFrom(sourceUnitClass)) {
					return -1;
				}
				
				throw new IllegalStateException("Classes "+sourceUnitClass.getName()+" and "+s.sourceUnitClass.getName()+" do not belong to the same inheritance hierarchy");
			}
			
			throw new IllegalArgumentException("Cannot compare "+this.getClass().getName()+" with "+o.getClass().getName());
		}
	}
	
	private Collection<SourceUnitClassEntry> sourceUnitClasses=new ArrayList<SourceUnitClassEntry>();
	
	private class LanguageElementClassEntry extends LanguageElementClassImpl implements Comparable {
		private Set<Number> requiredLevels=new HashSet<Number>();
		private Class languageElementClass;
		private Class contextClass;
		
		public LanguageElementClassEntry(ResultSet rs) throws SQLException {
			super(rs);
			try {
				languageElementClass=classLoader.loadClass(getClassName());
			} catch (ClassNotFoundException e) {
				throw new SQLExceptionEx("Cannot load language element class "+getClassName(), e);
			}
			
			if (!(LanguageElement.class.isAssignableFrom(languageElementClass) || LanguageElementFactory.class.isAssignableFrom(languageElementClass))) {
				throw new SQLException(languageElementClass.getName()+" must be a subclass of "+LanguageElement.class.getName()+" or implement "+LanguageElementFactory.class);
			}
	
			if (getContextClass()!=null) {
				try {
					contextClass=classLoader.loadClass(getContextClass());
				} catch (ClassNotFoundException e) {
					throw new SQLExceptionEx("Cannot load context class "+getContextClass(), e);
				}
			}
			
			factory.getEngine().getLanguageElementClassRequiredLevels(
					getId(), 
					requiredLevels,
					factory.toIntegerConverter);
		}
		
		boolean isCompatible(Class actualContextClass, Class targetClass, Collection loadedLevels, boolean noEnvironment) {			
			return !(getRequiresEnvironment() && noEnvironment) 
				&& (this.contextClass==null || (actualContextClass!=null && this.contextClass.isAssignableFrom(actualContextClass))) 
				&& loadedLevels.containsAll(requiredLevels)
				&& (targetClass==null || targetClass.isAssignableFrom(languageElementClass) || LanguageElementFactory.class.isAssignableFrom(languageElementClass));
		}

		public int compareTo(Object o) {
			if (o==this) {
				return 0;
			}
			
			if (o instanceof LanguageElementClassEntry) {
				LanguageElementClassEntry l = (LanguageElementClassEntry) o;
				if (languageElementClass.equals(l.languageElementClass)) {
					logger.warning("Two entries with the same class "+languageElementClass.getName());
					return 0;
				}
				
				if (languageElementClass.isAssignableFrom(l.languageElementClass)) {
					return 1;
				}
				
				if (l.languageElementClass.isAssignableFrom(languageElementClass)) {
					return -1;
				}
				
				String myContextClass = getContextClass();
				String hisContextClass = l.getContextClass();
				
				if (isBlank(myContextClass)) {
					if (!isBlank(hisContextClass)) {
						return 1;
					}
				} else {
					if (isBlank(hisContextClass)) {
						return -1;
					}
				}
				
				// TODO - Compare contexts hierarchy.
				
				logger.warning("Ambiguous language element class: "+languageElementClass.getName()+" and "+l.languageElementClass.getName());
				return languageElementClass.getName().compareTo(l.languageElementClass.getName()); // Alphabetially, though it is bad.
			}
			
			throw new IllegalArgumentException("Cannot compare "+this.getClass().getName()+" with "+o.getClass().getName());
		}
	}
	
	private static boolean isBlank(String str) {
		return str==null || str.trim().length()==0;
	}
	
	private TIntObjectHashMap languageElementClasses=new TIntObjectHashMap();

	SourceUnit instantiateSourceUnit(org.mesopotamia.sql.SourceUnit dbData, Scan scan) {		
		try {
			Collection<Number> loadLevels=factory.getEngine().getSourceUnitSuccessfulLoadLevels(
					dbData.getId(), 
					scan.getId(), 
					new HashSet<Number>(),
					factory.levelIdToIntegerConverter);
			
			SourceUnitClassEntry suce=null;
			Iterator<SourceUnitClassEntry> it=sourceUnitClasses.iterator();
			while (it.hasNext()) {
				SourceUnitClassEntry candidate=it.next();
				if (candidate.isCompatible(loadLevels) && (suce==null || candidate.compareTo(suce)<0)) {
					suce=candidate;
				}			
			}
			
			if (suce==null) {
				return new SourceUnit(dbData, scan, this, Collections.unmodifiableCollection(loadLevels));
			}
		
			return (SourceUnit) suce.sourceUnitClass
				.getConstructor(
					new Class[] {
							org.mesopotamia.sql.SourceUnit.class, 
							Scan.class, 
							RepositoryLanguage.class,
							Collection.class})
				.newInstance(
					new Object[] {
							dbData, 
							scan, 
							this,
							Collections.unmodifiableCollection(loadLevels)});
		} catch (Exception e) {
			throw new MesopotamiaRuntimeException("Cannot instantiate source unit "+dbData, e);
		}		
	}
	
	/**
	 * @param tokenTypeId token type id. 
	 * @return true if given token type id belongs to this language.
	 */
	boolean belongsTo(int tokenTypeId) {
		return id2type.containsKey(tokenTypeId);
	}		

	public LanguageElement instantiateLanguageElement(
			NodeData xData,
			Class contextClass, 
			Class targetClass, 
			Scan scan) {
		try {
			Collection<Number> loadLevels=scan.getSourceUnitLoadLevels(xData.getSourceUnitId());
			
			LanguageElementClassEntry lece=null;
			
			Collection lecc=(Collection) languageElementClasses.get(xData.getType());
			StringBuffer candidates=new StringBuffer("[");
			if (lecc!=null) {
				Iterator it=lecc.iterator();
				while (it.hasNext()) {
					LanguageElementClassEntry candidate=(LanguageElementClassEntry) it.next();
					candidates.append(candidate.getClassName());
					if (it.hasNext()) {
						candidates.append(", ");
					}
					if (candidate.isCompatible(contextClass, targetClass, loadLevels, scan.environment==null) && (lece==null || candidate.compareTo(lece)<0)) {
						lece=candidate;
					}			
				}
			}
			candidates.append("]");
						
			if (lece==null) {
				if (targetClass==null || targetClass.isAssignableFrom(SimpleLanguageElement.class)) {
					return new SimpleLanguageElement(xData, contextClass, scan, this, scan.environment);
				}
				
				
				String sourceUnitPath = "";
				try {
					sourceUnitPath=factory.getEngine().getSourceUnit(xData.getSourceUnitId()).getPath();
				} catch (SQLException e) {
					factory.consume(xData, e);
				}
				throw new MesopotamiaRuntimeException(
						"Cannot instantiate language element of type "
						+targetClass.getName()
						+" ("+xData.getLine()+":"+xData.getColumn()+")"
						+" from token type "
						+tokenTypeId2name(xData.getType())
						+" in "
						+contextClass
						+" context of source unit "
						+ xData.getSourceUnitId()
						+ " "
						+ sourceUnitPath
						+". None of candidates "
						+candidates.toString()
						+" is compatible with "+targetClass.getName());
			}
		
			Object ret = lece.languageElementClass
				.getConstructor(
					new Class[] {
							NodeData.class, 
							Class.class,
							Scan.class, 
							RepositoryLanguage.class, 
							Object.class})
				.newInstance(
					new Object[] {
							xData, 
							contextClass, 
							scan, 
							this, 
							scan.environment});
			
			if (ret instanceof LanguageElementFactory) {
				return ((LanguageElementFactory) ret).createLanguageElement(targetClass); 
			} else if (ret instanceof LanguageElement) {
				return (LanguageElement) ret;
			}
			
			throw new MesopotamiaRuntimeException(ret.getClass()+" is neither LangaugeElement nor LanguageElementFactory");
		} catch (Exception e) {
			throw new MesopotamiaRuntimeException("Cannot instantiate language element "+dbData, e);
		}		
	}
	
	boolean isScanDependent(int loadLevelId) {
		return ((LoaderEntry) loadersMap.get(loadLevelId)).isScanDependent();
	}

	/**
	 * Posts jobs to load source unit to subsequent levels.
	 */
	void load(
			final int scanId, 
			final int sourceUnitId, 
			final Source source, 
			final Object environment, 
			final SourceUnitLoadListener listener) {
		
		final TIntHashSet successfullyLoadedLevels=new TIntHashSet();
		final TIntHashSet loadedLevels=new TIntHashSet();
		try {
			factory.getEngine().processSourceUnitSuccessfulLoadLevels(
					sourceUnitId,
					scanId,
					new RowProcessor() {

						public boolean process(ResultSet rs) throws SQLException {							
							int levelId = rs.getInt("LEVEL_ID");
							loadedLevels.add(levelId);
							if (!rs.getBoolean("LOAD_FAILED")) {
								successfullyLoadedLevels.add(levelId);
							}
							return true;
						}
						
					});
		
			Iterator<LoaderEntry> it=loaders.iterator();
			while (it.hasNext()) {
				final LoaderEntry le=it.next();
				if (le.isCompatible(successfullyLoadedLevels) && !loadedLevels.contains(le.getId())) {
					
					if (le.getLoader() instanceof SourceLoader) {
						if (source!=null) {
							// TODO - make job public class, DOM Serializable in order to distribute.
							Runnable job = new Runnable() {
								public void run() {
									logger.fine("Loading to level "+le.getLevel());
									if (((SourceLoader) le.getLoader()).load(
											scanId, 
											sourceUnitId, 
											source, 
											environment)) {
											if (listener!=null) {
												listener.onLoad(scanId, sourceUnitId, le.getLevel());
											}
											load(scanId, sourceUnitId, source, environment, listener);
										}
								}
								
								public String toString() {
									return "Load job: "+le.getLoader().getClass()+" loading source "+source.getName()+" to source unit "+sourceUnitId;
								}
							};
							factory.process(job);
						} else {
							logger.severe("Source is null for loader "+le.getLevel());
						}
					} else if (le.getLoader() instanceof SourceUnitLoader) {
						// TODO - make job public class, DOM Serializable in order to distribute.
						Runnable job = new Runnable() {
							public void run() {
								logger.fine("Loading to level "+le.getLevel());
								
								if (((SourceUnitLoader) le.getLoader()).load(
										scanId, 
										sourceUnitId, 
										environment)) {
									if (listener!=null) {
										listener.onLoad(scanId, sourceUnitId, le.getLevel());
									}																						
									load(scanId, sourceUnitId, source, environment, listener);
								}
							}
							
							public String toString() {
								return "Load job: "+le.getLoader().getClass()+" loading source unit "+sourceUnitId;
							}
						};
						factory.process(job);
					} else {
						logger.severe("Loader is neither SourceLoader nor SourceUnitLoader");
					}
				}
			}
		} catch (SQLException e) {
			factory.consume(this, e);
		}
	}
	
	public String toString() {
		return "["+getClass().getName()+"] "+getName()+" "+getVersion();
	}

}
