/*
 * mesopotamia @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia;

import java.awt.event.WindowEvent;
import java.lang.ref.Reference;
import java.lang.ref.SoftReference;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.WindowConstants;
import javax.swing.tree.TreeNode;

import org.mesopotamia.sql.LoadError;
import org.mesopotamia.sql.MesopotamiaEngine;
import org.w3c.dom.Element;

import biz.hammurapi.config.Context;
import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.legacy.review.Signed;
import biz.hammurapi.render.HtmlRenderer;
import biz.hammurapi.swing.Browser;
import biz.hammurapi.swing.Visualizable;
import biz.hammurapi.util.Attributable;
import biz.hammurapi.util.VisitableBase;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * Represents source unit, e.g. java source file
 * @author Pavel Vlasov
 * @revision $Revision: 1.1 $
 */
public class SourceUnit extends VisitableBase implements DomSerializable, Signed, Attributable {
	Scan scan;
	RepositoryLanguage repoLanguage;
	private org.mesopotamia.sql.SourceUnit dbData;
	
	/**
	 * Map of level name -> Soft reference -> level data 
	 */
	private Map<String, Reference> levelData = new HashMap<String, Reference>();
	
	/**
	 * Set of level at which this unit is loaded.
	 */
	private Collection<Number> loadLevels = new ArrayList<Number>();

	/**
	 * Instances of SourceUnit shall be created only by
	 * repository
	 * @throws SQLException 
	 */
	protected SourceUnit(org.mesopotamia.sql.SourceUnit dbData, final Scan scan, RepositoryLanguage repoLanguage, Collection<Number> loadLevels) throws MesopotamiaException {
		this.repoLanguage=repoLanguage;
		this.scan=scan;
		this.dbData=dbData;
		
		Iterator<Number> it = loadLevels.iterator();
		while (it.hasNext()) {
			this.loadLevels.add(it.next());
			
		}		
		MesopotamiaEngine engine = scan.getRepository().getFactory().getEngine();		
		try {
			errors=Collections.unmodifiableCollection(engine.getLoadError(dbData.getId(), scan.getId(), new ArrayList<LoadError>()));
		} catch (SQLException e) {
			throw new MesopotamiaException("Could not load errors: "+e, e);
		}
	}
	
	static final List<Object> EMPTY_UNMODIFIABLE_LIST = Collections.unmodifiableList(new ArrayList());

	public int getId() {
		return dbData.getId();
	}

	public String getName() {
		return dbData.getName();
	}

	/**
	 * Namespace shall be retrieved from repository
	 * @return
	 */
	public Namespace getNamespace() {
		return scan.getNamespace(dbData.getNamespaceId());
	}

	public String getPath() {
		return dbData.getPath();
	}

	public String getDigest() {
		return dbData.getUnitDigest();
	}

	public Long getSize() {
		return dbData.getUnitSize();
	}
	
	public RepositoryLanguage getLanguage() {
		return repoLanguage;
	}

	public Scan getScan() {
		return scan;
	}

	public void toDom(Element holder) {
		holder.setAttribute("digest-algorithm", dbData.getDigestAlgorithm());
		holder.setAttribute("id", String.valueOf(dbData.getId()));
		holder.setAttribute("language", dbData.getLanguage());
		holder.setAttribute("language-version", dbData.getLanguageVersion());
		if (dbData.getLastModified()!=null) {
			holder.setAttribute("last-modified", dbData.getLastModified().toString());
		}
		holder.setAttribute("name", dbData.getName());
		holder.setAttribute("path", dbData.getPath());
		holder.setAttribute("repository", String.valueOf(dbData.getRepositoryId()));
		if (dbData.getUnitSize()!=null) {
			holder.setAttribute("size", dbData.getUnitSize().toString());
		}
		if (dbData.getUnitDigest()!=null) {
			holder.setAttribute("digest", dbData.getUnitDigest());
		}		
	}

	/**
	 * Returns signature, which is path.
	 */
	public String getSignature() {
		return getPath();
	}
	
	/**
	 * @return load levels of this source unit. Elements are of type org.mesopotamia.sql.SourceUnitSuccessfulLoadLevels
	 */
	public Collection<Number> getLoadLevels() {
		return loadLevels;
	}
	
	public String toString() {
		return "["+getClass().getName()+"] "+getPath();
	}
	
	private Collection<LoadError> errors;
	
	public Collection<LoadError> getErrors() {
		return errors;
	}
	
	/**
	 * @param levelName
	 * @return Load level data for given source unit. This method returns only 
	 * scan-independent data.
	 * @throws MesopotamiaException 
	 */
	public Object getLevelData(String levelName) {
		synchronized (levelData) {
			try {
				Reference<Object> ref = levelData.get(levelName);
				Object ret = ref==null ? null : ref.get();
				if (ret==null) {
					ret = getLanguage().getLoader(levelName).getData(getId(), null);
					ref = new SoftReference<Object>(ret);
					levelData.put(levelName, ref);
				}						
				return ret;
			} catch (MesopotamiaException e) {
				throw new MesopotamiaRuntimeException(e);
			}
		}
	}
	
	/**
	 * Shows source unit in browser. Use it for debugging.
	 */
	public void show() {
		Visualizable vs = (Visualizable) ConvertingService.convert(this, Visualizable.class); 
		TreeNode root = vs.toTreeNode(null, "Source unit");
		
		final Object monitor = new Object();
		
		Browser browser = new Browser("Source unit "+getId(), root) {
			
			protected void processWindowEvent(WindowEvent e) {
				super.processWindowEvent(e);
				if (e.getID()==WindowEvent.WINDOW_CLOSED) {
					synchronized (monitor) {
						monitor.notifyAll();
					}
				}
			}
		};
		
		browser.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		browser.setVisible(true);
		
		synchronized (monitor) {
			try {
				monitor.wait();
			} catch (InterruptedException ex) {
				// Ignore
			}
		}

	}
	
	private Map<Object, Object> attributes = new HashMap<Object, Object>();

	public Object getAttribute(Object key) {
		return attributes.get(key);
	}

	public Object removeAttribute(Object key) {
		return attributes.remove(key);
	}

	public void setAttribute(Object key, Object value) {
		attributes.put(key, value);		
	}
	
	/**
	 * Returns renderer to render this source unit to HTML.
	 * @param context Renderer configuration parameters.
	 * @return renderer or null if source unit does not support
	 * rendering.
	 */
	public HtmlRenderer getRenderer(Context context) {
		return null;
	}
}
