/*
 * mesopotamia @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.mesopotamia.MesopotamiaException;
import org.mesopotamia.MesopotamiaRuntimeException;
import org.onemind.jxp.CachedJxpPage;
import org.onemind.jxp.CachingPageSource;
import org.onemind.jxp.JxpContext;
import org.onemind.jxp.JxpPageSource;
import org.onemind.jxp.JxpProcessor;

import antlr.ASTFactory;
import antlr.RecognitionException;
import antlr.TokenStreamException;
import antlr.collections.AST;
import biz.hammurapi.util.CollectionVisitable;
import biz.hammurapi.util.VisitableBase;
import biz.hammurapi.util.Visitor;

public class BnfModel extends VisitableBase {
	
	private static final Logger logger = Logger.getLogger(BnfModel.class.getName());
	
	private Collection<Object> ruleDefinitions=new ArrayList<Object>();
	/**
	 * Keeps mapping subclass (String) -> superclass (RuleDefinition)
	 */
	Map<String, RuleDefinition> superClasses=new HashMap<String, RuleDefinition>();
	Map<String, Collection<RuleDefinition>> interfaces=new HashMap<String, Collection<RuleDefinition>>();
	Map<MatchPath, RuleDefinition> matchPath=new HashMap<MatchPath, RuleDefinition>();
	Map<String, RuleDefinition> rulesMap=new HashMap<String, RuleDefinition>(); // name -> rule
	private String packageName;
	private Class tokenTypesClass;
	
	void addInterface(String implementor, RuleDefinition theInterface) {
		Collection<RuleDefinition> interfaceCollection=interfaces.get(implementor);
		if (interfaceCollection==null) {
			interfaceCollection=new ArrayList<RuleDefinition>();
			interfaces.put(implementor, interfaceCollection);
		}
		interfaceCollection.add(theInterface);
		
		// Validation that no interface extends class.
		Iterator<Object> it=ruleDefinitions.iterator();
		while (it.hasNext()) {
			RuleDefinition rule=(RuleDefinition) it.next();
			if (rule.isInterface() && !(rule.getSuperRule()==null || rule.getSuperRule().isInterface())) {
				System.err.println("Interface rule "+rule.getName()+" extends class rule "+rule.getSuperClassName());
			}
		}		
	}
	
	public BnfModel(Reader r, Class tokenTypesClass, String rootLanguageElementClassName) throws MesopotamiaException {
		this.tokenTypesClass=tokenTypesClass;
		try {
			BnfLexer lexer=new BnfLexer(r);
			lexer.setTokenObjectClass(biz.hammurapi.antlr.Token.class.getName());
			BnfRecognizer parser=new BnfRecognizer(lexer);
			ASTFactory astFactory = new ASTFactory();
			astFactory.setASTNodeClass(biz.hammurapi.antlr.AST.class);
			parser.setASTFactory(astFactory);
			parser.grammar();
			AST ast=parser.getAST();
			
			PrintStream printStream = new PrintStream(new FileOutputStream("ast.txt"));
			((biz.hammurapi.antlr.AST) ast).print(parser.getTokenNames(), printStream, 0, true);
			printStream.close();
			
			Set<String> names=new HashSet<String>();
			for (AST node=ast; node!=null; node=node.getNextSibling()) {
				if (node.getType()==BnfTokenTypes.LITERAL_package) {
					StringBuffer sb=new StringBuffer();
					for (AST pnode=node.getFirstChild(); pnode!=null; pnode=pnode.getNextSibling()) {
						sb.append(pnode.getText());
					}
					packageName=sb.toString();
				} else if (names.add(node.getText())) {
					RuleDefinition ruleDefinition = new RuleDefinition(this, node, rootLanguageElementClassName);
					ruleDefinitions.add(ruleDefinition);
					rulesMap.put(ruleDefinition.getName(), ruleDefinition);
				} else {
					System.err.println("Rule already defined: "+node.getText());
				}
			}			
		} catch (TokenStreamException e) {
			throw new MesopotamiaException(e);
		} catch (IOException e) {
			throw new MesopotamiaException(e);
		} catch (RecognitionException e) {
			throw new MesopotamiaException(e);
		}
		
	}
	
	protected void acceptChildren(Visitor visitor) {
		new CollectionVisitable(ruleDefinitions, false).accept(visitor);
	}
	
	public void checkTokenName(String tokenName) throws MesopotamiaRuntimeException {
		if (tokenTypesClass!=null && !"*".equals(tokenName)) {
			try {
				tokenTypesClass.getField(tokenName);
			} catch (SecurityException e) {
				throw new MesopotamiaRuntimeException("Field "+tokenName+" cannot be accessed", e);
			} catch (NoSuchFieldException e) {
				throw new MesopotamiaRuntimeException("Invalid token name: "+tokenName);
			}
		}
	}
	

	/**
	 * @param args
	 */
	public static void main(final String[] args) throws Exception {
		System.out.println("Arguments: <bnf file> <token types class> <root language element class> <template directory> <output directory>");
		if (args.length!=5) {
			System.err.println("Invalid number of arguments");
			return;
		}
		
		BnfModel model = new BnfModel(new FileReader(args[0]), Class.forName(args[1]), args[2]);
		
		final File templateDir = new File(args[3]);
		
		final JxpPageSource pageSource = new CachingPageSource() {
			
			protected boolean isExpired(CachedJxpPage page) {
				return false;
			}

			protected boolean hasStream(String pageName) {
				return new File(templateDir, pageName+".jxp").isFile();
			}

			protected InputStream loadStream(CachedJxpPage page) throws IOException {
				File pageFile = new File(templateDir, page.getName()+".jxp");
				page.setPageTimepstamp(pageFile.lastModified());
				return new FileInputStream(pageFile);
			}
		};
		
		final File outputDir = new File(args[4]);
		
		model.accept(new Visitor() {

			public boolean visit(Object target) {
				logger.info("Processing: "+target);
				Map<String, Object> context = new HashMap<String, Object>();
				JxpContext jxpContext = new JxpContext(pageSource, context);
				JxpProcessor processor = new JxpProcessor(jxpContext);
				try {
					if (target instanceof RuleDefinition) {
						context.put("rule", target);
						RuleDefinition rd = (RuleDefinition) target;
						String templateName;
						if (rd.isFactory()) {
							templateName = "factory";
						} else if (rd.isInterface()) {
							templateName = "interface";
						} else {
							templateName = "class";
						}
						FileWriter writer = new FileWriter(new File(outputDir, rd.getName()+".java"));
						try {
							processor.process(templateName, writer);
						} finally {
							writer.close();
						}
						return false;
					} else if (target instanceof BnfModel) {
						context.put("model", target);
						FileWriter writer = new FileWriter(new File(outputDir, "LanguageModel.java"));
						try {
							processor.process("model", writer);
						} finally {
							writer.close();
						}	
					}
				} catch (Exception e) {
					logger.log(Level.WARNING, "Processing of "+target+" failed: "+e, e);
				}
				return true;
			}
			
		});				
	}

	public String getPackage() {
		return packageName;
	}
}
