/*
  * mesopotamia @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
  */
package org.mesopotamia.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import javax.swing.tree.TreeNode;

import org.mesopotamia.LanguageElement;
import org.mesopotamia.MesopotamiaNode;
import org.mesopotamia.Scan;
import org.mesopotamia.SourceUnit;
import org.mesopotamia.SyntaxTree;

import biz.hammurapi.convert.Converter;
import biz.hammurapi.swing.LazyTreeNode;
import biz.hammurapi.swing.LazyTreeNodeTableVisualizable;
import biz.hammurapi.swing.Visualizable;


/**
 * @author Pavel Vlasov
 *
 * @version $Revision: 1.1 $
 */
public class MesopotamiaVisualizer {
	
	/**
	 * Invokes method and returns non-null value if the method result shall be mounted to the tree.
	 * @param method
	 * @return true if return value shall be mounted to tree, false if it shall be displayed in the table.
	 * @throws InvocationTargetException 
	 * @throws IllegalAccessException 
	 * @throws  
	 */
	private Object getNode(Method method, Object obj) throws IllegalAccessException, InvocationTargetException {
		if (LanguageElement.class.isAssignableFrom(method.getDeclaringClass())) {
			if (((LanguageElement) obj).showInTree(method)) {
				Object returnValue = method.invoke(obj);
				if ((LanguageElement.class.isInstance(returnValue) || List.class.isInstance(returnValue))) {
					return returnValue;
				}
			}
		} else if (SourceUnit.class.isAssignableFrom(method.getDeclaringClass())) {
			if ("getLanguageElements".equals(method.getName()) || "getSyntaxTree".equals(method.getName())) {
				return method.invoke(obj);
			}
		} else if (Scan.class.isAssignableFrom(method.getDeclaringClass())) {
			if ("getSourceUnits".equals(method.getName())) {
				return method.invoke(obj);
			}
		} else if (SyntaxTree.class.isAssignableFrom(method.getDeclaringClass())) {
			if ("getRoots".equals(method.getName())) {
				return method.invoke(obj);
			}
		} else if (MesopotamiaNode.class.isAssignableFrom(method.getDeclaringClass())) {
			if ("getChildren".equals(method.getName())) {
				return method.invoke(obj);
			}
		}
		return null;
	}
	
	public Visualizable convert(final LanguageElement le, final Converter master) {
		return new Visualizable() {

			public TreeNode toTreeNode(TreeNode parent, String title) {
				return _toTreeNode(parent, title, le, master);
			}
			
		};
	}

	public Visualizable convert(final Scan sc, final Converter master) {
		return new Visualizable() {

			public TreeNode toTreeNode(TreeNode parent, String title) {
				return _toTreeNode(parent, title, sc, master);
			}
			
		};
	}

	public Visualizable convert(final SourceUnit su, final Converter master) {
		return new Visualizable() {

			public TreeNode toTreeNode(TreeNode parent, String title) {
				return _toTreeNode(parent, title, su, master);
			}
			
		};
	}

	public Visualizable convert(final SyntaxTree st, final Converter master) {
		return new Visualizable() {

			public TreeNode toTreeNode(TreeNode parent, String title) {
				return _toTreeNode(parent, title, st, master);
			}
			
		};
	}

	public Visualizable convert(final MesopotamiaNode mn, final Converter master) {
		return new Visualizable() {

			public TreeNode toTreeNode(TreeNode parent, String title) {
				return _toTreeNode(parent, title, mn, master);
			}
			
		};
	}
	
	private TreeNode _toTreeNode(final TreeNode parent, final String title, final Object obj, final Converter master) {
		return new LazyTreeNodeTableVisualizable(parent, title) {

			public String toString() {
				if (title!=null && title.trim().length()>0) {
					return title;
				}
				
				String ret = obj.getClass().getName();
				int idx = ret.lastIndexOf(".");
				return idx==-1 ? ret : ret.substring(idx+1);
			}
			
			@Override
			protected List<TreeNode> loadChildren() {
				List<TreeNode> ret = new ArrayList<TreeNode>();
				if (obj instanceof SyntaxTree) {
					SyntaxTree st = (SyntaxTree) obj;
					Iterator<MesopotamiaNode> it = st.getRoots().iterator();
					while (it.hasNext()) {
						MesopotamiaNode mn = it.next();
						Visualizable vs = (Visualizable) master.convert(mn, Visualizable.class, null);
						ret.add(vs.toTreeNode(this, mn.getTypeName() + " " + (mn.getText()==null ? "" : mn.getText())));
					}
				} else if (obj instanceof MesopotamiaNode) {
					MesopotamiaNode mNode = (MesopotamiaNode) obj;
					Iterator<MesopotamiaNode> it = mNode.getChildren().iterator();
					while (it.hasNext()) {
						MesopotamiaNode mn = it.next();
						Visualizable vs = (Visualizable) master.convert(mn, Visualizable.class, null);
						ret.add(vs.toTreeNode(this, mn.getTypeName() + " " + (mn.getText()==null ? "" : mn.getText())));
					}					
				} else {				
					if (obj instanceof LanguageElement) {
						MesopotamiaNode mNode = ((LanguageElement) obj).getNode();
						final Collection<MesopotamiaNode> nChildren = mNode.getChildren();
						if (!nChildren.isEmpty()) {
							ret.add(new LazyTreeNode(this, "Syntax tree") {

								@Override
								protected List<TreeNode> loadChildren() {
									List<TreeNode> ret = new ArrayList<TreeNode>();
									Iterator<MesopotamiaNode> it = nChildren.iterator();
									while (it.hasNext()) {
										MesopotamiaNode mn = it.next();
										Visualizable vs = (Visualizable) master.convert(mn, Visualizable.class, null);
										ret.add(vs.toTreeNode(this, mn.getTypeName() + " " + (mn.getText()==null ? "" : mn.getText())));						
									}			
									return ret;
								}
								
							});
							
						}
					}
					
				    Method[] methods = obj.getClass().getMethods();
				    for (int i=0; i<methods.length; i++) {
				    	// getXXX() methods. Object and LanguageElement classes are not included.
				    	if (!(methods[i].getDeclaringClass().equals(Object.class))
				    			&& methods[i].getName().startsWith("get") 
				    			&& !Modifier.isStatic(methods[i].getModifiers())
				    			&& !methods[i].isSynthetic()
				    			&& methods[i].getParameterTypes().length==0) {
				    		try {
								Object value=getNode(methods[i], obj);
								if (value!=null && (!(value instanceof Collection) || !((Collection) value).isEmpty())) {
									Visualizable vs = (Visualizable) master.convert(value, Visualizable.class, null);
									ret.add(vs.toTreeNode(this, methods[i].getName().substring("get".length())));
								}
							} catch (Exception e) {
								e.printStackTrace();								
							}
				    	}					    						    	
				    }
				}
				
				return ret;
			}

			public TableModel toTable() {
				int rowCount=2;
				DefaultTableModel tm=new DefaultTableModel(rowCount,4);			
				tm.setColumnIdentifiers(new String[] {"Property", "Declared type", "Runtime type", "Value"});
				
			    Class<? extends Object> beanClass=obj.getClass();
			    
				tm.setValueAt("this", 0, 0);
				tm.setValueAt(beanClass.getName(), 0, 2);
				tm.setValueAt(obj, 0, 3);
				
				tm.setValueAt("Hash code", 1, 0);
				tm.setValueAt("int", 1, 1);
				tm.setValueAt("int", 1, 2);
				tm.setValueAt(Integer.toString(obj.hashCode(), Character.MAX_RADIX), 1, 3);
				
			    Method[] methods = beanClass.getMethods();
			    List<Method> ml = new ArrayList<Method>();
			    for (int i=0; i<methods.length; i++) {
			    	ml.add(methods[i]);
			    }
			    
			    Collections.sort(ml, new Comparator<Method>() {

					public int compare(Method m0, Method m1) {
						return m0.getName().compareTo(m1.getName());
					}
			    	
			    });
			    
			    Iterator<Method> it = ml.iterator();
			    while (it.hasNext()) {
			    	Method method = it.next();
			    	// getXXX() methods. Object.getClass() is not included.
			    	if (!(method.getDeclaringClass().equals(Object.class)) 
			    			&& method.getName().startsWith("get") 
			    			&& method.getParameterTypes().length==0) {
			    		try {
			    			if (getNode(method, obj)==null) {
								Object value=method.invoke(obj);
								if (value!=null) {
									tm.setRowCount(++rowCount);
									int idx = rowCount-1;
									tm.setValueAt(method.getName().substring(3), idx, 0);
									tm.setValueAt(method.getReturnType().getName(), idx, 1);
									tm.setValueAt(value.getClass().getName(), idx, 2);
									tm.setValueAt(value, idx, 3);
								}
			    			}
						} catch (Exception e) {
							e.printStackTrace();								
						}
			    	}					    						    	
			    }
			    			    
				return tm;
			}
			
		};
		
	}
}
